/*
 Subroutines for data acquisition and processing
 
 Copyright (c) 2008 Dimtel, Inc., All Rights Reserved

 $Id: subDataAcq.c,v 1.1.1.1 2009/06/02 20:25:59 dim Exp $
 $Date: 2009/06/02 20:25:59 $
 $Author: dim $
 $Revision: 1.1.1.1 $
 $Log: subDataAcq.c,v $
 Revision 1.1.1.1  2009/06/02 20:25:59  dim
 Initial import


*/

#include <dbDefs.h>
#include <registryFunction.h>
#include <subRecord.h>
#include <epicsExport.h>
#include <dbAccess.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

/* FFT code definitions */
#include <complex.h>
#include <fftw3.h>

/* Local stuff */
#include "genSubRecord.h"
#include "BxB.h"
#include "cmath.h"
#include "spf.h"
#include "dacq.h"

extern acq_data snapshot;
extern pthread_mutex_t snapshot_mutex;

extern int USB_Debug;

proc_data vectors;

typedef long (*processMethod)(subRecord *precord);

static long subDataAcqInit(subRecord *psub, processMethod process)
{
  psub->val = 0;
  return 0;
}

static void dacq_proc_bxb(void)
{
  int k, m, diff, bidx, Navg;
  long turns, spt;
  double rms0max, foo;

  /* FFTW structures */
  static fftw_complex *fftout;
  static fftw_plan p;
  
  /* Timing structures */
  struct timeval t1, t2;

  if (USB_Debug == 1)
    gettimeofday(&t1, NULL);

  spt = vectors.bxb.samples_per_turn;

  /* Compute the number of turns */
  turns = ADC_LENGTH/spt;

  if (vectors.bxb.first)
  {
    /* FFT output/input */
    fftout  = fftw_malloc(sizeof(fftw_complex)*(turns/2 + 1));
    /* FFT computation - use tmp as input data vector */
    p = fftw_plan_dft_r2c_1d(turns, vectors.bxb.tmp, fftout, FFTW_ESTIMATE);
    vectors.bxb.first = 0;
  }

  /* Count enabled bunches */
  for (k = 0, Navg = 0; k < spt; k++)
    Navg += (int)vectors.bxb.mask[k];

  if (Navg == 0) Navg = 1;
      
  bzero(vectors.bxb.mag, (turns/2+1)*sizeof(double));
  for (k = 0, bidx = 0; k < spt; k++)
  {
    if (vectors.bxb.mask[k] == 1) /* Enabled for averaging */
    {
      /* Fill in the input data */
      for (m = 0; m < turns; m++)
        vectors.bxb.tmp[m] = (double)vectors.bxb.adc[k+m*spt];
  
      /* Transform it */    
      fftw_execute(p);
    
      /* Accumulate squares of absolute values in mag */
      for (m = 0; m <= turns/2; m++)
        if (bidx == 0)
          vectors.bxb.mag[m] = creal_loc(fftout[m]*
                                conj_loc(fftout[m]))/Navg/turns;
        else
          vectors.bxb.mag[m] += creal_loc(fftout[m]*
                               conj_loc(fftout[m]))/Navg/turns;
      bidx++;
    }
  }
  
  /* Convert to dB */
  foo = 0.0;
  for (m = 0; m <= turns/2; m++)
  {
    vectors.bxb.mag[m] = 10*log10(vectors.bxb.mag[m]*4+1e-10);
    foo += vectors.bxb.mag[m]/(turns/2+1);
  }
  /* Set the DC value to the average */
  vectors.bxb.mag[0] = foo;

  /* Do the averaging */
  if (!vectors.bxb.scale_update)
    for (m = 0; m <= turns/2; m++)
      vectors.bxb.mag[m] = vectors.alpha*vectors.bxb.mag_old[m] +
                           (1.0 - vectors.alpha)*vectors.bxb.mag[m];
  memcpy(vectors.bxb.mag_old, vectors.bxb.mag, (turns/2+1)*sizeof(double));
      
  /* Copy data to double vector */
  for (m = 0; m < turns*spt; m++)
    vectors.bxb.data[m] = (double)vectors.bxb.adc[m];
  
  /* Compute the average vector */
  spf_mean_trans(vectors.bxb.data, vectors.bxb.mean, spt, turns);
  
  /* Compute the RMS vector */
  spf_rms_dc_trans(vectors.bxb.data, vectors.bxb.mean, vectors.bxb.rms, spt, turns);
  
  /* Find the element with the largest RMS and copy it out */
  k = spf_max(vectors.bxb.rms, &rms0max, spt);

#if 0  
  if (rms0max > 60)
    vectors.bxb.stop = 1;
#endif

  for (m = 0; m < turns; m++)
    vectors.bxb.max_rms[m] = (long)(vectors.bxb.adc[k + m*spt] -
                             vectors.bxb.mean[k]);
  
  if (vectors.bxb.scale_update)
  {
    for (m = 0; m <= turns/2; m++)
      vectors.bxb.freq[m] = (float)m / vectors.bxb.us_per_sample / 
                      spt * 1000.0 / turns;
    for (m = 0; m < turns; m++)
      vectors.bxb.tsc[m] = (float)(m*(vectors.bxb.us_per_sample/1000)*spt);
    vectors.bxb.scale_update = 0;
  }
  if (USB_Debug == 1)
  {
    gettimeofday(&t2, NULL);
    diff = deltaT(&t1, &t2);
    printf("Elapsed time (DataProcThread): %.3f ms\n", ((double)diff)/1000.0);
  }
}

static void dacq_proc_raw(void)
{
  int m;

  /* FFTW structures */
  static fftw_complex *fftout;
  static fftw_plan p;
  
  if (vectors.raw.first)
  {
    /* FFT output/input */
    fftout  = fftw_malloc(sizeof(fftw_complex)*(ADC_LENGTH/2 + 1));
    /* FFT computation - use vectors.raw.tmp as input data vector */
    p = fftw_plan_dft_r2c_1d(ADC_LENGTH, vectors.raw.tmp, fftout, FFTW_ESTIMATE);
    vectors.raw.first = 0;
  }

  /* Fill in the input data */
  for (m = 0; m < ADC_LENGTH; m++)
    vectors.raw.tmp[m] = (double)vectors.raw.adc[m];
          
  /* Transform it */    
  fftw_execute(p);
    
  /* Save squares of absolute values in mag */
  for (m = 0; m <= ADC_LENGTH/2; m++)
    vectors.raw.mag[m] = creal_loc(fftout[m]*conj_loc(fftout[m]))/ADC_LENGTH;
  
  /* Extract square root of mag and scale */
  for (m = 0; m <= ADC_LENGTH/2; m++)
    vectors.raw.mag[m] = 10.0*log10(vectors.raw.mag[m]+1e-10);
  vectors.raw.mag[0] = vectors.raw.mag[1]; /* Get rid of DC offset */
  vectors.raw.mag[0] = 0;

  /* Do the averaging */
  if (!vectors.raw.scale_update)
    for (m = 0; m <= ADC_LENGTH/2; m++)
      vectors.raw.mag[m] = vectors.alpha*vectors.raw.mag_old[m] +
                       (1.0 - vectors.alpha)*vectors.raw.mag[m];
                             
  for (m = 0; m <= ADC_LENGTH/2; m++)
    vectors.raw.mag_old[m] = vectors.raw.mag[m];
      
  if (vectors.raw.scale_update)
  {
    for (m = 0; m <= ADC_LENGTH/2; m++)
      vectors.raw.freq[m] = (float)m/vectors.raw.us_per_sample/ADC_LENGTH;

    vectors.raw.scale_update = 0;
  }

}

static long subDataAcqProc(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = Sampling frequency
              B = Averaging factor
              C = Acquisition mode, 0 - raw, 1 - bxb
              D = Harmonic number
              E = RF harmonic used for sampling
              F = Spectral averaging mask
   * Outputs:
   *          A = ADC data
   *          B = raw mode ADC spectrum
   *          C = raw frequency vector
   *          D = raw scan disable
   *          E = bxb mean
   *          F = bxb rms
   *          G = bxb spectrum
    *          H = bxb frequency
   *          I = bxb max rms
   *          J = bxb tsc
   *          K = bxb raw data
   *          L = bxb scan disable
   */
  
  int turns, hnum, rf_harm;
  static int first = 1;
  static double fsamp;
  double fs;

  /* Timing structures */
  int diff;
  struct timeval t1, t2;

  if (USB_Debug == 1)
    gettimeofday(&t1, NULL);
        
  if (USB_Debug == 2)
  {
        printf("Record %s called subDataAcqbeamproc(%p)\n",
               pgsub->name, (void*) pgsub);
  }

  fs   = *(double *)pgsub->a;
  hnum = *(long *)pgsub->d;
  rf_harm = *(long *)pgsub->e;
  
  if (hnum == 0)
    return 0;

  turns = ADC_LENGTH/hnum;
  
  if (first)
  {
    /* First-time init */
    vectors.raw.us_per_sample = 1/fs;
    vectors.bxb.us_per_sample = rf_harm/fs;
    vectors.bxb.samples_per_turn = hnum;
    fsamp = fs;
    vectors.raw.scale_update = vectors.bxb.scale_update = 1;
    vectors.raw.first = vectors.bxb.first = 1;
    vectors.mode = *(char *)pgsub->c;
    vectors.bxb.stop = 0;
    first = 0;
    
    /* Raw mode output vectors */
    vectors.raw.adc     = (long *)pgsub->vala;
    vectors.raw.mag     = (double *)pgsub->valb;
    vectors.raw.freq    = (float *)pgsub->valc;
    
    /* Allocate BxB memory */
    vectors.bxb.mag_old = (double *)malloc((turns/2+1)*sizeof(double));
    vectors.bxb.tmp     = (double *)malloc(turns*sizeof(double));

    /* BxB mode input/output vectors */
    vectors.bxb.mask    = (char *)  pgsub->f;
    vectors.bxb.mean    = (double *)pgsub->vale;
    vectors.bxb.rms     = (double *)pgsub->valf;
    vectors.bxb.mag     = (double *)pgsub->valg;
    vectors.bxb.freq    = (float *) pgsub->valh;
    vectors.bxb.max_rms = (int *)   pgsub->vali;
    vectors.bxb.tsc     = (float *) pgsub->valj;
    vectors.bxb.data    = (double *)pgsub->valk;
  }
  
  /* Try to get the lock */
  if (pthread_mutex_trylock(&snapshot_mutex) == 0)
  {
    if (snapshot.new != 0) /* New data is available */
    {
      if (vectors.mode != *(char *)pgsub->c)
      {
        vectors.mode = *(char *)pgsub->c;
          
        /* Make sure when we switch back the averaging is reset */
        if (vectors.mode == DACQ_MODE_RAW)
          vectors.bxb.scale_update = 1;
        else
          vectors.raw.scale_update = 1;
      }
      if (vectors.mode == DACQ_MODE_RAW)
        memcpy(vectors.raw.adc, snapshot.adc, ADC_LENGTH*sizeof(long));
      else
        memcpy(vectors.bxb.adc, snapshot.adc+2,
               turns*vectors.bxb.samples_per_turn*sizeof(long));
#if 0
        for (k = 2; k < turns*vectors.bxb.samples_per_turn+2; k++)
        {
          m = (k-2) % vectors.bxb.samples_per_turn; /* Channel number */
          int idx = m*turns + (k-2)/vectors.bxb.samples_per_turn;
          vectors.bxb.adc[idx] = snapshot.adc[k];
        }
#endif      
      vectors.alpha = 1.0 - 1.0 / *(double *)pgsub->b;
      if (fsamp != fs)
      {
        vectors.raw.us_per_sample = 1/fs;
        vectors.bxb.us_per_sample = rf_harm/fs;
        fsamp                     = fs;
        vectors.raw.scale_update  = 1;
      }
      snapshot.new = 0;
      
      if (vectors.mode == DACQ_MODE_RAW)
      {
        dacq_proc_raw();
        *(char *)pgsub->vald = 0;
        *(char *)pgsub->vall = 1;
        vectors.bxb.stop = 0;
      }
      else
      {
        if (vectors.bxb.stop == 0)
          dacq_proc_bxb();
        *(char *)pgsub->vald = 1;
        *(char *)pgsub->vall = 0;
      }
    }
    else
      /* Disable updates without new data */
      *(char *)pgsub->vald = *(char *)pgsub->vall = 1;
    pthread_mutex_unlock(&snapshot_mutex);
  }    
  
  if (USB_Debug == 1)
  {
    gettimeofday(&t2, NULL);
    diff = deltaT(&t1, &t2);
    printf("Elapsed time (SubDataAcqProc): %.3f ms\n", ((double)diff)/1000.0);
  }

  return 0;
}

static long subDataAcqbeamstat(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = mean vector
   *          B = rms vector
   *          C = frequency vector
   *          D = spectrum vector
   *          E = Spectral search lower frequency (kHz)
   *          H = Spectral search upper frequency (kHz)
   *          I = Spectral search flag (0 - min, 1 - max)
   *          F = Harmonic number
   *          G = Number of samples
   * Outputs: A = mean around the turn
   *          B = peak-to-peak mean
   *          C = RMS around the turn
   *          D = max RMS
   *          E = Marker amplitude (dB)
   *          F = Marker frequency (kHz)
   *          G = Marker amplitude (vector of 2)
   *          H = Marker frequency (vector of 2)
   */
  
  int j, k, m, flag=0;
  long turns, frlen;
  double mean_max, mean_min, center, base, marker_amp[2], marker_freq[2];
  double amp_pp, peak;
  float *rms, *mean, *freq, *spec;
  static long samples_per_turn = 0;

  if (USB_Debug == 2)
        printf("Record %s called subDataAcqbeamstat(%p)\n",
               pgsub->name, (void*) pgsub);

  if (samples_per_turn == 0)
  {
    if (*(long *)pgsub->f == 0)
      return 0;
    else
      samples_per_turn = *(long *)pgsub->f;
  }

  turns = ADC_LENGTH/samples_per_turn;
  frlen = turns/2 + 1;
  mean = (float *) pgsub->a;
  rms  = (float *) pgsub->b;
  freq = (float *) pgsub->c;
  spec = (float *) pgsub->d;
  
  /* Compute average mean, put it in E */
  spf_mean_f(mean, (double *)pgsub->vala, 1, samples_per_turn);
  /* Compute RMS, put it in G */
  spf_rms_f(rms, (double *)pgsub->valc, 1, samples_per_turn);
  /* Compute max RMS, put it in H */
  spf_max_f(rms, (double *)pgsub->vald, samples_per_turn);

  /* Find max and min of the mean, separate along the centerline */
  spf_max_f(mean, &mean_max, samples_per_turn);
  spf_min_f(mean, &mean_min, samples_per_turn);
  center = (mean_max + mean_min) / 2;
  
  base = peak = 0.0;
  j = m = 0;
  for (k = 0; k < samples_per_turn; k ++)
    if (mean[k] > center)
    {
      peak += mean[k];
      j++;
    }
    else
    {
      base += mean[k];
      m++;
    }
  if (j != 0)
  {
    peak /= j;
    base /= m;
    /* Peak-to-peak amplitude (average orbit) */
    amp_pp = peak - base;
  }
  else
    amp_pp = 0.0;
    
  *(double *)pgsub->valb = amp_pp;
  
  /* Marker search */
  for (k = frlen - 1; k >= 0; k--)
    if (freq[k] <= *(double *)pgsub->e)
      break;
    else if (freq[k] <= *(double *)pgsub->h)
    {
      if (flag == 0)
      {
        marker_amp[0] = spec[k];
        marker_freq[0] = freq[k];
        flag = 1;
      }
      else if (*(char *)pgsub->i == 1)
      {
        if (spec[k] > marker_amp[0])
        {
          marker_amp[0] = spec[k];
          marker_freq[0] = freq[k];
        }
      }
      else
        if (spec[k] < marker_amp[0])
        {
          marker_amp[0] = spec[k];
          marker_freq[0] = freq[k];
        }
    }
  *(double *)pgsub->vale = marker_amp[0];
  *(double *)pgsub->valf = marker_freq[0];
  
  marker_amp[1] = marker_amp[0]; marker_freq[1] = marker_freq[0];
  memcpy(pgsub->valg, marker_amp, 2*sizeof(double));
  memcpy(pgsub->valh, marker_freq, 2*sizeof(double));

  return 0;
}

static long subDataAcqPattern(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = Harmonic number
   *          B = Bunch pattern string
   *
   * Outputs: A = Pattern mask
   */

  long hnum;

  /* Read input data */
  hnum         = *(long *)  pgsub->a;

  if (hnum == 0) return 0;
  
  /* Set up the bunch mask. */
  patternConvert(pgsub->b, pgsub->vala, hnum);
  
  return 0;
}

static registryFunctionRef subDataAcqRef[] = {
  {"subDataAcqInit", (REGISTRYFUNCTION)subDataAcqInit},
  {"subDataAcqProc", (REGISTRYFUNCTION)subDataAcqProc},
  {"subDataAcqPattern", (REGISTRYFUNCTION)subDataAcqPattern},
  {"subDataAcqbeamstat", (REGISTRYFUNCTION)subDataAcqbeamstat}
};


static void subDataAcqRegistrar(void)
{
  registryFunctionRefAdd(subDataAcqRef, NELEMENTS(subDataAcqRef));
}
epicsExportRegistrar(subDataAcqRegistrar);
