/*
 FIR filter control, 5 coefficient sets
 
 Copyright (c) 2009 Dimtel, Inc., All Rights Reserved

 $Id$
 $Date$
 $Author$
 $Revision$
 $Log$

*/

#include   <stdio.h>
#include   <string.h>
#include   <math.h>
#include   <complex.h>

#include   <dbDefs.h>
#include   <registryFunction.h>
#include   <subRecord.h>
#include   <epicsExport.h>
#include   <dbAccess.h>

#include   "genSubRecord.h"
#include   "cmath.h"
#include   "map.h"
#include   "fir.h"

typedef long (*processMethod)(subRecord *precord);

static long subFilterInit(genSubRecord *pgsub)
{
  pgsub->val = 0;
  return 0;
}

/* Copy coefficient vector */
static long subCopyCoeff(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = Set to load
   *          B = Vector to load
   *          C = ID
   *          D = Comment
   *          E = Shift gain
   * Outputs: A = Coefficient set vector
   *          B = Shift gain
   *          C = Comment
   */
  int m, set_to_load;
  long nRequest;

  /* Copy coefficient set on init */
  nRequest = MAXFIRLEN;
  m = dbGetLink(&pgsub->outa, DBR_LONG, pgsub->vala, NULL, &nRequest);
  for (m = nRequest; m < MAXFIRLEN; m++)
    ((long *)pgsub->vala)[m] = 0;

  /* Copy shift gain */
  nRequest = 1;
  dbGetLink(&pgsub->outb, DBR_LONG, pgsub->valb, NULL, &nRequest);

  /* Copy comment */
  nRequest = 1;
  dbGetLink(&pgsub->outc, DBR_STRING, pgsub->valc, NULL, &nRequest);

  set_to_load = *(long *)pgsub->a;

  if (set_to_load != *(long *)pgsub->c)
    return 0;
      
  nRequest = MAXFIRLEN;
  dbGetLink(&pgsub->inpb, DBR_LONG, pgsub->vala, NULL, &nRequest);
  bzero((long *)pgsub->vala + nRequest, (MAXFIRLEN - nRequest)*sizeof(long));
  
  *(long *)pgsub->valb = *(long *)pgsub->e;

  memcpy(pgsub->valc, pgsub->d, 40);
         
  return 0;
}

/* Generate FIR filter coefficients */
static long subFilterGen(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = gain (0 to 1)
   *          B = phase (-180 to 180)
   *          C = freq (0 to 1)
   *          D = taps (1 to 16)
   * Outputs: E = Coefficient vector
   */
  char  comment[41];
  int coeff[MAXFIRLEN];
  int k, taps;
  double gain, phase, freq, temp, cf[MAXFIRLEN];
  
  gain  = *((double *)pgsub->a);
  phase = *((double *)pgsub->b);
  freq  = *((double *)pgsub->c);
  taps  = *((long *)pgsub->d);
  
  /* Clip to MAXFIRLEN to prevent wild segfaults */
  if (taps > MAXFIRLEN)
    taps = MAXFIRLEN;

#if 0
  printf("subFilterGen called with gain=%0.2f;phase=%3.1f;freq=%0.2f;taps=%d\n",
         gain, phase, freq, taps);
#endif
  for (k = 0, temp = 0.0; k<taps; k++)
  {
    cf[k] = gain*sin(2.0*M_PI*freq*k + phase*M_PI/180.0);
    temp += cf[k];
  }
  temp /= taps;
  
  if (freq != 0.0)
  {
    for (k = 0; k < taps; k++)
      cf[k] -= temp;
    
    /* Find the largest coefficient */
    for (k = 0, temp = 0.0; k < taps; k++)
      if (fabs(cf[k]) > temp)
        temp = fabs(cf[k]);

    /* Normalize to gain */
    if (temp != 0.0)
    {
      for (k = 0; k < taps; k++)
        cf[k] *= gain/temp;
    }
  }
  for (k = 0; k < taps; k++)
    coeff[k] = (int)(FIR_COEFF_FS*cf[k]);
  
  for (; k < MAXFIRLEN; k++)
    coeff[k] = 0;

  snprintf(comment, 41, "Gain=%0.2f;Phase=%0.1f;Freq=%0.2f;Taps=%d",
          gain, phase, freq, taps);

  memcpy(pgsub->vala, coeff, MAXFIRLEN*sizeof(int));
  memcpy(pgsub->valb, comment, 40);
    
  return 0;
}

/* Compute FIR filter TF */
static long subFilterTF(genSubRecord *pgsub)
{
  /*
   * Inputs:  A = Coefficient vector
   *          B = Fractional tune
   *          C = Shift gain
   * Outputs: A = Magnitude vector
   *          B = Phase vector
   *          C = Frequency vector
   *          D = Gain at tune frequency
   *          E = Phase at tune frequency
   *          F = Tune frequency (2 long vector)
   */
  int points = pgsub->nova, k, m;
  long *coeff;
  long shift_gain;
  double tune, freq, f_start, f_stop;
  double complex z, H;
  
  coeff      = pgsub->a;
  tune       = *(double *) pgsub->b;
  shift_gain = 1<<(*(long *) pgsub->c);
  f_start = 0;
  f_stop  = 0.5;
  
  for (k = 0; k < points; k++)
  {
    freq = f_start + k*(f_stop - f_start)/(points - 1);
    z = cexp_loc(-I*2*M_PI*freq);
    for (m = MAXFIRLEN-1, H = 0.0; m >= 0; m--)
      H = coeff[m] + H * z;
      
    H = H / FIR_COEFF_FS * shift_gain;
    *((double *)pgsub->vala + k) = 20.0*log10(cabs(H)+3.0e-3);
    *((double *)pgsub->valb + k) = 180*carg_loc(H)/M_PI;
    *((double *)pgsub->valc + k) = freq;
  }

  z = cexp_loc(-I*2*M_PI*tune);
  for (m = MAXFIRLEN - 1, H = 0.0; m >= 0; m--)
    H = coeff[m] + H * z;

  *(double *)pgsub->vald = 20.0*log10(cabs(shift_gain*H/FIR_COEFF_FS)+3.0e-3);
  *(double *)pgsub->vale = 180*carg_loc(H)/M_PI;
  *((double *)pgsub->vald + 1) = *(double *)pgsub->vald;
  *((double *)pgsub->vale + 1) = *(double *)pgsub->vale;
  *(double *)pgsub->valf = tune;
  *((double *)pgsub->valf + 1) = tune;

  return 0;
}

static registryFunctionRef subFilterRef[] = {
  {"subFilterInit", (REGISTRYFUNCTION)subFilterInit},
  {"subCopyCoeff", (REGISTRYFUNCTION)subCopyCoeff},
  {"subFilterGen", (REGISTRYFUNCTION)subFilterGen},
  {"subFilterTF", (REGISTRYFUNCTION)subFilterTF}
};

static void subFilterRegistrar(void)
{
  registryFunctionRefAdd(subFilterRef, NELEMENTS(subFilterRef));
}
epicsExportRegistrar(subFilterRegistrar);
