/*
 Miscellaneous subroutines
 
 Copyright (c) 2006 Dimtel, Inc., All Rights Reserved

 $Id: subMisc.c,v 1.2 2009/06/07 17:02:03 dim Exp $
 $Date: 2009/06/07 17:02:03 $
 $Author: dim $
 $Revision: 1.2 $
 $Log: subMisc.c,v $
 Revision 1.2  2009/06/07 17:02:03  dim
 Added FMAX fanout function for updating DRVH/HOPR values on frequency controls

 Revision 1.1.1.1  2009/06/02 20:25:59  dim
 Initial import

 Revision 1.3  2007/12/30 15:27:06  dim
 Phase servo, drive pattern generation.

 Revision 1.2  2007/10/28 07:02:40  dim
 Drive subsystem, spectral marker, test interface.

 Revision 1.1  2007/03/14 08:34:56  dim
 IOC code updates for multithreaded DSP, new features


*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>

#include <dbDefs.h>
#include <registryFunction.h>
#include <subRecord.h>
#include <epicsExport.h>
#include <dbAccess.h>
#include <math.h>

#include "genSubRecord.h"
#include "BxB.h"
#include "dacq.h"

typedef long (*processMethod)(subRecord *precord);

static long subMiscInit(subRecord *psub, processMethod process)
{
  psub->val = 0;
  return 0;
}

static long subIPaddr(genSubRecord *pgsub)
{
  /*
   * Outputs: A = IP address string
   */
  char host_name[256], ip_addr[16];
  struct hostent *host;
  
  gethostname(host_name, 256);
  host = gethostbyname(host_name);
  
  sprintf(ip_addr, "%hhu.%hhu.%hhu.%hhu", host->h_addr[0], host->h_addr[1],
          host->h_addr[2], host->h_addr[3]);

  memcpy(pgsub->vala, ip_addr, strlen(ip_addr)+1);
  
  return 0;
}

/* Fan out an input through channel access */
static long subMiscFanout(subRecord *psub)
{
  /*
   * Outputs: A = Input value
   *          B = Number of output links
   *          C = Offset
   *          D = Links...
   */
  int n = (int)psub->b;
  double val = psub->a + psub->c;
    
  if (n > 0)
    dbPutLink(&psub->inpd, DBR_DOUBLE, &val, 1);
  if (n > 1)
    dbPutLink(&psub->inpe, DBR_DOUBLE, &val, 1);
  if (n > 2)
    dbPutLink(&psub->inpf, DBR_DOUBLE, &val, 1);
  if (n > 3)
    dbPutLink(&psub->inpg, DBR_DOUBLE, &val, 1);
  if (n > 4)
    dbPutLink(&psub->inph, DBR_DOUBLE, &val, 1);
  if (n > 5)
    dbPutLink(&psub->inpi, DBR_DOUBLE, &val, 1);
  if (n > 6)
    dbPutLink(&psub->inpj, DBR_DOUBLE, &val, 1);
  if (n > 7)
    dbPutLink(&psub->inpk, DBR_DOUBLE, &val, 1);
  if (n > 8)
    dbPutLink(&psub->inpl, DBR_DOUBLE, &val, 1);
  
  return 0;
}

static long subMiscConfigParse(subRecord *psub)
{
  /*
   * Inputs:  A = Config register
   * Outputs: B = Harmonic number
   *          C = Gateware revision
   *
  */
  ulong cfg = (ulong)psub->a, harmonic;
  uchar rev, rev_maj, rev_min;
  double revision;
   
  harmonic = cfg & 0x7f;
  rev      = (cfg >> 16) & 0xff;
  rev_maj  = rev/16;
  rev_min  = rev % 16;
  revision = (double)rev_maj + (double)rev_min/100;
   
  dbPutLink(&psub->inpb, DBR_LONG, &harmonic, 1);
  dbPutLink(&psub->inpc, DBR_DOUBLE, &revision, 1);

  return 0;
}

static long subMiscDS1822ROM(genSubRecord *pgsub)
{
  /*
   * Input:   A = ROM array (2x32)
   * Outputs: A = Serial number string
   *          B = Checksum status (0 - OK)
   *          C = Device ID
   *          D = Future scan disable
   */
  unsigned char rom[8], crc;
  unsigned char table[256] = 
  {0, 94, 188, 226, 97, 63, 221, 131, 194, 156, 126, 32, 163, 253, 31, 65,
   157, 195, 33, 127, 252, 162, 64, 30, 95, 1, 227, 189, 62, 96, 130, 220,
   35, 125, 159, 193, 66, 28, 254, 160, 225, 191, 93, 3, 128, 222, 60, 98,
   190, 224, 2, 92, 223, 129, 99, 61, 124, 34, 192, 158, 29, 67, 161, 255,
   70, 24, 250, 164, 39, 121, 155, 197, 132, 218, 56, 102, 229, 187, 89, 7,
   219, 133, 103, 57, 186, 228, 6, 88, 25, 71, 165, 251, 120, 38, 196, 154,
   101, 59, 217, 135, 4, 90, 184, 230, 167, 249, 27, 69, 198, 152, 122, 36,
   248, 166, 68, 26, 153, 199, 37, 123, 58, 100, 134, 216, 91, 5, 231, 185,
   140, 210, 48, 110, 237, 179, 81, 15, 78, 16, 242, 172, 47, 113, 147, 205,
   17, 79, 173, 243, 112, 46, 204, 146, 211, 141, 111, 49, 178, 236, 14, 80,
   175, 241, 19, 77, 206, 144, 114, 44, 109, 51, 209, 143, 12, 82, 176, 238,
   50, 108, 142, 208, 83, 13, 239, 177, 240, 174, 76, 18, 145, 207, 45, 115,
   202, 148, 118, 40, 171, 245, 23, 73, 8, 86, 180, 234, 105, 55, 213, 139,
   87, 9, 235, 181, 54, 104, 138, 212, 149, 203, 41, 119, 244, 170, 72, 22,
   233, 183, 85, 11, 136, 214, 52, 106, 43, 117, 151, 201, 74, 20, 246, 168,
   116, 42, 200, 150, 21, 75, 169, 247, 182, 232, 10, 84, 215, 137, 107, 53};
  unsigned *inrom = (unsigned *)pgsub->a, k;
  unsigned long long serial;
  
  /* Convert in host-independent way */
  rom[0] = (inrom[0]      ) & 0xff;
  rom[1] = (inrom[0] >>  8) & 0xff;
  rom[2] = (inrom[0] >> 16) & 0xff;
  rom[3] = (inrom[0] >> 24) & 0xff;
  rom[4] = (inrom[1]      ) & 0xff;
  rom[5] = (inrom[1] >>  8) & 0xff;
  rom[6] = (inrom[1] >> 16) & 0xff;
  rom[7] = (inrom[1] >> 24) & 0xff;

  for (k = 0, crc = 0; k < 8; k++)
    crc = table[crc ^ rom[k]];
    
  *(char *) pgsub->valb = crc ? 1 : 0;
  *(char *) pgsub->valc = rom[0];
  *(char *) pgsub->vald = 1; /* Only first pass needed */

  serial = (unsigned long long)((inrom[0] >> 8) & 0xffffff) | 
           ((unsigned long long)(inrom[1] & 0xffffff) << 24);
  sprintf(pgsub->vala, "%012llx", serial);
/*  printf("DS1822: 0x%08x, 0x%08x, %llx\n", inrom[0], inrom[1], serial);*/ 

  return 0;
}

static registryFunctionRef subMiscRef[] = {
  {"subMiscInit", (REGISTRYFUNCTION)subMiscInit},
  {"subIPaddr", (REGISTRYFUNCTION)subIPaddr},
  {"subMiscFanout", (REGISTRYFUNCTION)subMiscFanout},
  {"subMiscDS1822ROM", (REGISTRYFUNCTION)subMiscDS1822ROM},
  {"subMiscConfigParse", (REGISTRYFUNCTION)subMiscConfigParse}
};

static void subMiscRegistrar(void)
{
  registryFunctionRefAdd(subMiscRef, NELEMENTS(subMiscRef));
}
epicsExportRegistrar(subMiscRegistrar);
